/*
 *  check.c - checking functions
 *  Copyright (c)2002 - 2004 Andrej  OM1CW
 *  Copyright (c)2005 Mike Karas OM4AA <mike.om4aa@gmail.com>
 *  Copyright (c)2009 Ladislav Vaiz OK1ZIA <ok1zia@nagano.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "osdep.h"
#include "check.h"
#include "main.h"
#include "debug.h"
#include "call.h"
#include "compare.h"
#include "cdwdb.h"
#include "load.h"


int good_qso(struct QSO_INFO *q)
{
	DEBUG_FUNC("good_qso");

//    printf("\ngood_qso(%s %s %s %s %d)\n", q->s_call, q->ok_call, q->ok_rst, q->r_rst, q->invalid_rsts);
    if (q->pass >=100) return 0; // NIL, DUPE
    if (q->ok_call != q->r_call) return 0;
    if (valid_rst(q->ok_rst)){
        if (strcasecmp(q->ok_rst,q->r_rst)!=0) return 0;
    }
    if (checknr && q->ok_nr != q->r_nr) return 0;
    if (!q->r_call->loc) return 0; 
    if (checknr && !q->r_nr) return 0; 
    if (q->wwl_outside) return 0;
    if (q->invalid_rsts) return 0;

    // wihtout penboth only old checks
    if (penboth == 0) return 1;

    // check from other side ---------------------------------
    q = q->ok_qso;
    if (!q) return 1; // cannot check, no log available
    
    if (q->pass >=100) return 0; // NIL, DUPE
    if (q->ok_call != q->r_call) return 0;
    if (valid_rst(q->ok_rst)){
        if (strcasecmp(q->ok_rst,q->r_rst)!=0) return 0;
    }
    if (checknr && q->ok_nr != q->r_nr) return 0;
    if (!q->r_call->loc) return 0; 
    if (checknr && !q->r_nr) return 0; 
    if (q->wwl_outside) return 0;
    if (q->invalid_rsts) return 0;

    return 1;

}

void make_xchk2(uint8_t band, uint8_t pass, uint8_t mask, char *foundby)
{
	DEBUG_FUNC("make_xchk2");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q, *r;
	uint32_t p;
//	int entries = 0;
	uint8_t found;

	// printf("+ Xcheck pass #%i: ", pass);

	printf(".");

	for(c = first_call[band]; c != NULL; c = c->next) {
		for (q = c->first_qso; q != NULL; q = q->next) {
            if (q->pass) continue;
            for (r = q->ok_call->first_qso; r != NULL; r = r->next) {
                found = 0;
                if (r->s_call == q->s_call) continue; // we cannot make himself 
                if (r->pass) continue;
                if (r->phase != q->phase) continue;
//                                    printf("57: ok_call=%s %s\n", q->r_call->call, r->s_call);

                p = 0;
                if (!strcmp(r->ok_call->call, q->s_call->call))
                    p += 64;	// he rcvd good CALL
                // if (r->s_call == q->ok_call) p+=64;
                // //ja som prijal dobre CALL

//                                printf("%s %d  %s %d\n", r->r_rst, r->invalid_rsts, q->s_rst, q->invalid_rsts);
                if (strcasecmp(r->r_rst,q->s_rst)==0)
                    p += 32;	// he rcvd good RST

//                                printf("%s %d  %s %d\n", r->s_rst, r->invalid_rsts, q->r_rst, q->invalid_rsts);
                if (strcasecmp(r->s_rst,q->r_rst)==0)
                    p += 16;	// I rcvd good RST

                if (checknr && r->r_nr == q->s_nr)
                    p += 8;	// he rcvd good NR
                if (checknr && r->s_nr == q->r_nr)
                    p += 4;	// I rcvd good NR
                if (r->ok_call->loc == q->s_call->loc)
                    p += 2;	// he rcvd good WWL
                if (r->s_call->loc == q->ok_call->loc)
                    p += 1;	// I rcvd good WWL

//                                printf("%x %x %x\n", p, mask, p & mask);
                if ((p & mask) == mask) {
                    found = 1;
//									entries += 2;
                    if ((p & 1) && r->ok_call->refs >= 20000){
                       if (verbose_mode >= 4) printf("skip: %s %d -> %s %d because call is in force.wwl\n", r->ok_call->call, r->ok_call->refs, q->s_call->call, q->s_call->refs);
                       break;
                    }
                    q->ok_qso = r;
                    r->ok_qso = q;
                    q->foundby = foundby;
                    r->foundby = foundby;
                    if (verbose_mode >= 4){
                       printf("\nmask=%d foundby=%s\n", mask, foundby);
//                       printf("87: q=%p q->ok_qso=%s r=%p r->ok_qso=%s\n", 
//                            q->s_call, q->ok_qso->s_call, r->s_call, r->ok_qso->s_call);
                       printf("89: %s %d -> %s %d\n", r->ok_call->call, r->ok_call->refs, q->s_call->call, q->s_call->refs);
//                     printf("90: ok_call=%s %s\n", q->r_call->call, q->ok_qso->s_call);
//                     printf("91: ok_call=%s %s\n", r->r_call->call, r->ok_qso->s_call);
                    }
                    q->pass = pass;
                    r->pass = pass;
                    break;
                }
            }
		}
	}

	// printf("DONE...%d%% of entries\n", entries*100/all_entries);
}

void make_not_in_log(uint8_t band)
{
	DEBUG_FUNC("make_not_in_log");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q, *r;
	uint32_t /* entries=0, */ lost = 0;
	uint8_t found;

	// printf("+ Xcheck pass #%i: ", pass);

	printf(".");

	while (c) {
		q = c->first_qso;
		if (q) {
			while (q) {
				if (!q->pass) {
					r = q->ok_call->first_qso;
					if (r) {
						found = 0;
						while (r) {
							if (!strcmp(r->ok_call->call, c->call)) {
								found = 1;
								break;
							}
                            if (checknr && q->ok_nr == r->s_nr){
                                q->nil_qso = r;
                            }
							r = r->next;
						}

						if (!found) {
							lost++;
							q->pass = 100;
						}
					}

				}

				q = q->next;
			}
		}
		c = c->next;
	}

	// printf("DONE...%d%% of entries\n", entries*100/all_entries);
}

/* check and correct bad calls */
void make_bad_call(uint8_t band)
{
	DEBUG_FUNC("make_bad_call");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q;
	char text[SIZE_CALL + 1];
	int bad_call = 0;

	while (c) {
		q = c->first_qso;
		while (q) {
			if (q->r_call->refs <= 3) {
//                printf("176: %s q->r_call->refs=%d\n", q->r_call->call, q->r_call->refs);
				if ((bad_call = found_in_bad_calls(q->r_call->call)) != -1) {
					strncpy(text, bad_calls[bad_call].correct, SIZE_CALL);
					if (verbose_mode > 1){
						fprintf(inf, "%s correcting BADCALL: %s to %s!\n", c->call, q->r_call->call, text);
					    //printf("%s correcting BADCALL: %s to %s!\n", c->call, q->r_call->call, text);
                    }
				} else{
					strncpy(text, q->r_call->call, SIZE_CALL);
                }
				if (strcmp(text, q->r_call->call)){
					q->ok_call = add_call(text, q->ok_call->loc, band);
                    if (!q->foundby) q->foundby = "bad call";
                    if (verbose_mode >= 4) printf("check223: add %s\n", q->ok_call->call);
                }
			}
			q = q->next;
		}
		c = c->next;
	}
}

void make_xchk_res(uint8_t band)
{
	DEBUG_FUNC("make_xchk_res");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q;

	printf(".");

	// printf("+ Xcheck result: ");

	while (c) {
		q = c->first_qso;
		while (q) {
			if (q->ok_qso) {
//                printf("204: ok_call=%s %s\n", q->r_call->call, q->ok_qso->s_call);
				q->ok_call = q->ok_qso->s_call;
                if (!q->foundby) q->foundby = "xchk res";
                if (verbose_mode >= 4) printf("check249: ok_call=%s\n", q->ok_call->call);
				strcpy(q->ok_rst, q->ok_qso->s_rst);
				q->ok_nr = q->ok_qso->s_nr;
			}
			q = q->next;
		}
		c = c->next;
	}

	// printf("DONE\n");
}

void make_dupe(uint8_t band)
{
	DEBUG_FUNC("make_dupe");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q, *q2;

	// printf("+ DUPE check: ");

	printf(".");

	while (c) {
		q = c->first_qso;
		// if (q) printf(".");
		while (q) {
			if (good_qso(q) && !q->dupe) {
				q2 = q->next;
				while (q2) {
					if (compare_calls2(q->ok_call->call, q2->ok_call->call) && q->phase == q2->phase){
//                        printf("compare_calls2('%s', '%s')\n", q->ok_call->call, q2->ok_call->call);
						q2->pass = 101;
                    }
					q2 = q2->next;
				}
			}
			q = q->next;
		}
		c = c->next;
	}

	// printf("OK\n");
}

void make_outside_wwl(uint8_t band)
{
#ifdef HAVE_GLIB2
	DEBUG_FUNC("make_outside_wwl");
	
	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q;
	char loc[20];
	int have_log;
	FILE *fd;
	
	if (!dw)
		return;

	printf(".");
	fd = open_repfile(band, 1);
	fprintf(fd, "\n\nMaking check of outside WWLs:\n");
	while (c) {
		q = c->first_qso;
		have_log = 0;
		while (q) {
            int wwlok=wwl_is_ok_by_call(dw, num2loc(q->ok_call->loc, loc), q->ok_call->call);
            if (!wwlok){
                char *d = rindex(q->ok_call->call, '/');
                if (d != NULL && strcasecmp(d, "/MM") == 0) {
                    wwlok = 1;
				    if (verbose_mode > 1 && c->categ > 0) {
				        fprintf(fd, "%s: %s(%s) is outside of the country", c->call, q->ok_call->call, num2loc(q->ok_call->loc, loc));
						fprintf(fd, ", but is /MM - NOT repaired!\n");
                    }
                }
            }
            if (!wwlok) {
				if (q->ok_call->first_qso)
					have_log = 1;
				else
					q->wwl_outside = 1;
				if (verbose_mode > 1 && c->categ > 0) {
					fprintf(fd, "%s: %s(%s) is outside of the country", c->call, q->ok_call->call, num2loc(q->ok_call->loc, loc));
					if (have_log)
						fprintf(fd, ", but have log - NOT repaired!");
					fprintf(fd, "\n");
				}
			}
            if (wwlok==2){
                eprintf("%s: can't find dxcc for call %s\n", c->filepath, q->ok_call->call);
            }
			q = q->next;
		}
		c = c->next;
	}
	close_repfile(fd);
#endif
}

/* according to OK2ZI:
   Platny report je na SSB cislo v intervalu 31-39, 41-49, 51-59
   Na CW to je trochu horsi  311-319 , 411-419 atd az do 519 - 599
   Na VKV muze byt posledni cilo v trimistnem nahrazeno pismenem
   S - scatter, A - aurora, M multipath. 
   */

int valid_rst(const char *rst){
    char c;
    if (!rst) return 0;
    switch(strlen(rst)){
        case 2:
            if (rst[0]<'3' || rst[0]>'5') return 0;
            if (rst[1]<'1' || rst[1]>'9') return 0;
            return 1;
        case 3:
            if (rst[0]<'3' || rst[0]>'5') return 0;
            if (rst[1]<'1' || rst[1]>'9') return 0;
            c=toupper(rst[2]);
            if (c=='S') return 1;
            if (c=='A') return 1;
            if (c=='M') return 1;
            if (rst[2]<'1' || rst[2]>'9') return 0;
            return 1;
    }
    return 0;
}

void make_valid_rst(uint8_t band)
{
#ifdef HAVE_GLIB2
	DEBUG_FUNC("make_valid_rst");
	
	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q;
//	char loc[20];
	int have_log;
	FILE *fd;
	
	if (!dw)
		return;

	printf(".");
	fd = open_repfile(band, 1);
	fprintf(fd, "\n\nMaking check of valid RSTs:\n");
	while (c) {
		q = c->first_qso;
		have_log = 0;
		while (q) {
            q->invalid_rsts = !valid_rst(q->s_rst);
            if (q->invalid_rsts){
				if (verbose_mode > 1 && c->categ > 0) {
					fprintf(fd, "%s: %s invalid report %s\n", c->call, q->ok_call->call, q->s_rst);
				}
            }
			q = q->next;
		}
		c = c->next;
	}
	close_repfile(fd);
#endif
}

void make_uniq()
{
	DEBUG_FUNC("make_uniq");

    int band;
	struct CALL_INFO *c;
	struct QSO_INFO *q;
    FILE *f;
    char loc[7];
    char text[BUFF_SIZE + 1];

    sprintf(text, "%s%s%s", cfg_epath, DIR_DELIMITER, "lx.uni");
    f = fopen(text, "wt");
    if (!f){
        eprintf("Can't open %s\n", text);
        return;
    } 


    for (band = 1; band < BAND_NUM; band++) {
        if (!first_call[band]) continue;

        if (band >= BAND_GHZ) {
            if (band >= BAND_GHZ_ZERO)
                sprintf(text, "%2.0f GHz", band2freq[band]);
            else
                sprintf(text, "%2.1f GHz", band2freq[band]);
        } else
            sprintf(text, "%2.0f MHz", band2freq[band]);

        fprintf(f, "Band: %s\n", text);

        // clear refs
        for (c = first_call[band]; c != NULL; c = c->next ) {
            for (q = c->first_qso; q != NULL; q = q->next ) {
               q->r_call->refs = 0;
            }
        }
        // compute refs
        for (c = first_call[band]; c != NULL; c = c->next ) {
            for (q = c->first_qso; q != NULL; q = q->next ) {
                q->r_call->refs++;
            }
        }

        // output
        for (c = first_call[band]; c != NULL; c = c->next ) {
            if (c->first_qso == NULL) continue;
            if (c->categ == 0) continue;

            int nl = 0;
            for (q = c->first_qso; q != NULL; q = q->next ) {
                if (q->ok_call->refs != 1) continue;         // not unique
                if (q->ok_call->first_qso != NULL) continue; // have log, ignore

                fprintf(f, "%-12s ", c->call);
                fprintf(f, "%03d: ", q->s_nr);

                num2loc(q->ok_call->loc, loc);
                fprintf(f, "%-12s %s", q->ok_call->call, loc);

                fprintf(f, "\n");
                nl = 1;
            }
            if (nl) fprintf(f, "\n");
        }
        fprintf(f, "\n\n");
    }

    fclose(f);
}

void make_dump(){
	DEBUG_FUNC("make_uniq");

    int band;
	struct CALL_INFO *c, *c2;
	struct QSO_INFO *q;
    FILE *f;
    char loc[7];
    char text[BUFF_SIZE + 1];

    sprintf(text, "%s%s%s", cfg_epath, DIR_DELIMITER, "0lx.dump");
    f = fopen(text, "wt");
    if (!f){
        eprintf("Can't open %s\n", text);
        return;
    } 


    for (band = 1; band < BAND_NUM; band++) {
        if (!first_call[band]) continue;

        if (band >= BAND_GHZ) {
            if (band >= BAND_GHZ_ZERO)
                sprintf(text, "%2.0f GHz", band2freq[band]);
            else
                sprintf(text, "%2.1f GHz", band2freq[band]);
        } else
            sprintf(text, "%2.0f MHz", band2freq[band]);

        fprintf(f, "Band: %s\n", text);

        // clear refs
        for (c = first_call[band]; c != NULL; c = c->next ) {
            num2loc(c->loc, loc);
            fprintf(f, "%-12s %6s %d\n", c->call, loc, c->refs);

            for (q = c->first_qso; q != NULL; q = q->next ) {
                if (q->ok_call->first_qso != NULL) continue; // have log, ignore

                fprintf(f, "%-12s ", c->call);
                fprintf(f, "%03d: ", q->s_nr);

                c2 = q->s_call;
                num2loc(c2->loc, loc);
                fprintf(f, "\tsent(%s %s %03d %s)", c2->call, q->s_rst, q->s_nr, loc);

                c2 = q->r_call;
                num2loc(c2->loc, loc);
                fprintf(f, "\trcvd(%s %s %03d %s)", c2->call, q->r_rst, q->r_nr, loc);

                if (q->r_call != q->ok_call){
                    c2 = q->ok_call;
                    num2loc(c2->loc, loc);
                    fprintf(f, "\tok(%s %s %03d %s)", c2->call, q->ok_rst, q->ok_nr, loc);
                }

                fprintf(f, "\n");
            }
            fprintf(f, "\n");
        }
    }

}
